"""
gallerize.cli
~~~~~~~~~~~~~

:Copyright: 2007-2021 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

import argparse
from pathlib import Path
import sys

from . import VERSION
from .main import create_gallery, generate_gallery, handle_duplicate_filenames
from .models import Config, Dimension


ARGS_DEFAULT_MAX_IMAGE_SIZE: str = '1024x1024'
ARGS_DEFAULT_MAX_THUMBNAIL_SIZE: str = '120x120'


def parse_dimension_arg(value: str) -> Dimension:
    """Validate a dimension value."""
    try:
        return Dimension(*map(int, value.split('x', 1)))
    except ValueError:
        raise argparse.ArgumentTypeError(f'invalid dimension value: {value!r}')


def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser(
        usage='%(prog)s [options] <target path> <image> [image] ...'
    )

    parser.add_argument(
        '-c',
        '--captions',
        dest='captions',
        action='store_true',
        default=False,
        help='read image captions from text files ("<IMAGE_NAME>.txt")',
    )

    parser.add_argument(
        '--html-only',
        dest='html_only',
        action='store_true',
        default=False,
        help='only (re-)generate HTML, do not process images',
    )

    parser.add_argument(
        '--optimize-images',
        dest='optimize_images',
        action='store_true',
        default=False,
        help='optimize images to reduce size and remove metadata',
    )

    parser.add_argument(
        '--no-resize',
        dest='resize_images',
        action='store_false',
        default=True,
        help='do not resize images, just copy them',
    )

    parser.add_argument(
        '-s',
        '--size',
        dest='max_image_size',
        type=parse_dimension_arg,
        default=ARGS_DEFAULT_MAX_IMAGE_SIZE,
        help=f'set maximum image size [default: {ARGS_DEFAULT_MAX_IMAGE_SIZE}]',
    )

    parser.add_argument(
        '-t',
        '--thumbnail-size',
        dest='max_thumbnail_size',
        type=parse_dimension_arg,
        default=ARGS_DEFAULT_MAX_THUMBNAIL_SIZE,
        help=f'set maximum thumbnail size [default: {ARGS_DEFAULT_MAX_THUMBNAIL_SIZE}]',
    )

    parser.add_argument(
        '--title', dest='title', help='set gallery title on the website'
    )

    parser.add_argument(
        '--version', action='version', version=f'gallerize {VERSION}'
    )

    # First positional argument.
    parser.add_argument('destination_path', type=Path)

    # Remaining positional arguments (at least one), as a list.
    parser.add_argument('full_image_filenames', type=Path, nargs='+')

    return parser.parse_args()


def main():
    try:
        args = parse_args()

        handle_duplicate_filenames(args.full_image_filenames)

        gallery = create_gallery(
            title=args.title,
            destination_path=args.destination_path,
            max_image_size=args.max_image_size,
            max_thumbnail_size=args.max_thumbnail_size,
            full_image_filenames=args.full_image_filenames,
        )

        config = Config(
            html_only=args.html_only,
            resize_images=args.resize_images,
            optimize_images=args.optimize_images,
        )

        generate_gallery(gallery, config)
    except KeyboardInterrupt:
        sys.exit('Ctrl-C pressed, aborting.')



"""
gallerize.images
~~~~~~~~~~~~~~~~

Image processing

:Copyright: 2007-2021 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from pathlib import Path
import shutil
import subprocess

from .logging import debug
from .models import Dimension, Gallery, Image


def generate_images(gallery: Gallery, resize: bool, optimize: bool) -> None:
    """Generate images for the gallery."""
    for image in gallery.images:
        _generate_image(
            image,
            gallery.destination_path,
            resize,
            gallery.max_image_size,
            optimize,
        )
        _generate_thumbnail(
            image,
            gallery.destination_path,
            gallery.max_thumbnail_size,
            optimize,
        )


def _generate_image(
    image: Image,
    destination_path: Path,
    resize: bool,
    max_size: Dimension,
    optimize: bool,
) -> None:
    """Create an (optionally resized) copy of an image."""
    destination_filename = destination_path / image.filename
    if resize:
        # Resize image.
        debug('Resizing image "{}" ...', image.full_filename)
        _resize_image(image.full_filename, destination_filename, max_size)
    else:
        # Copy image.
        debug('Copying image "{}" ...', image.full_filename)
        shutil.copy(image.full_filename, destination_filename)

    if optimize:
        _optimize_image(destination_filename)


def _generate_thumbnail(
    image: Image, destination_path: Path, max_size: Dimension, optimize: bool
) -> None:
    """Create a small preview image for an image."""
    debug('Creating thumbnail "{}" ...', image.thumbnail_filename)
    destination_filename = destination_path / image.thumbnail_filename
    _resize_image(image.full_filename, destination_filename, max_size)

    if optimize:
        _optimize_image(destination_filename)


def _resize_image(
    src_filename: Path, dst_filename: Path, max_dimension: Dimension
) -> None:
    """Create a resized (and antialiased) version of an image."""
    dimension_str = f'{max_dimension.width:d}x{max_dimension.height:d}'
    cmd = [
        'convert',
        '-resize',
        dimension_str,
        str(src_filename),
        str(dst_filename),
    ]
    subprocess.check_call(cmd)


def _optimize_image(filename: Path) -> None:
    """Optimize image and strip EXIF and other non-important metadata."""
    cmd = ['jpegoptim', '--strip-all', str(filename)]
    subprocess.check_call(cmd)



"""
:Copyright: 2007-2021 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

import pytest

from gallerize.cli import parse_dimension_arg


@pytest.mark.parametrize(
    ('arg_value', 'expected_width', 'expected_height'),
    [
        ('480x640', 480, 640),
        ('1280x960', 1280, 960),
    ],
)
def test_parse_dimension_arg(arg_value, expected_width, expected_height):
    actual = parse_dimension_arg(arg_value)

    assert actual.width == expected_width
    assert actual.height == expected_height



"""
gallerize.html
~~~~~~~~~~~~~~

HTML rendering

:Copyright: 2007-2021 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from __future__ import annotations
from pathlib import Path
from typing import Any

from jinja2 import Environment, PackageLoader

from .files import write_file
from .models import Gallery, Image


TEMPLATE_EXTENSION: str = '.html'
OUTPUT_HTML_EXTENSION: str = '.html'

TEMPLATE_ENVIRONMENT: Environment = Environment(
    autoescape=True,
    loader=PackageLoader(__package__, 'templates'),
    trim_blocks=False,
)


def render_html_pages(gallery: Gallery) -> None:
    for image in gallery.images:
        _render_html_page(gallery, image)
    _render_html_index_page(gallery)


def _render_html_page(gallery: Gallery, image: Image) -> None:
    """Create an HTML document for a single image."""
    context = {'gallery': gallery, 'image': image}
    _render_html_to_file(
        'view', context, gallery.destination_path, image.page_name
    )


def _render_html_index_page(gallery: Gallery) -> None:
    """Create an HTML document of thumbnails that link to single image
    documents.
    """
    context = {'gallery': gallery}
    _render_html_to_file('index', context, gallery.destination_path, 'index')


def _render_html_to_file(
    template_name: str, context: dict[str, Any], path: Path, page_name: str
) -> None:
    """Render the template and write the result to the given file."""
    context['url_for_page'] = lambda page: page + OUTPUT_HTML_EXTENSION
    html = _render_template(template_name + TEMPLATE_EXTENSION, **context)

    filename = path / (page_name + OUTPUT_HTML_EXTENSION)
    write_file(filename, html)


def _render_template(template_filename: str, **context: dict[str, Any]) -> str:
    return TEMPLATE_ENVIRONMENT.get_template(template_filename).render(context)



"""
gallerize.logging
~~~~~~~~~~~~~~~~~

:Copyright: 2007-2021 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""


from typing import Any


def debug(message: str, *args: Any) -> None:
    print(message.format(*args))



"""
:Copyright: 2007-2021 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

import pytest

from gallerize.main import window


@pytest.mark.parametrize(
    ('iterable', 'expected'),
    [
        (
            [1],
            [
                (None, 1, None),
            ],
        ),
        (
            [1, 2],
            [
                (None, 1, 2),
                (1, 2, None),
            ],
        ),
        (
            [1, 2, 3],
            [
                (None, 1, 2),
                (1, 2, 3),
                (2, 3, None),
            ],
        ),
        (
            [1, 2, 3, 4],
            [
                (None, 1, 2),
                (1, 2, 3),
                (2, 3, 4),
                (3, 4, None),
            ],
        ),
    ],
)
def test_window(iterable, expected):
    assert list(window(iterable)) == expected



"""
gallerize.main
~~~~~~~~~~~~~~

:Copyright: 2007-2021 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from __future__ import annotations
import dataclasses
from itertools import islice
from pathlib import Path
import shutil
import sys
from typing import Iterable, Iterator, Optional

from .files import find_duplicate_filenames, read_first_line, write_file
from .html import render_html_pages
from .images import generate_images
from .logging import debug
from .models import Config, Dimension, Gallery, Image


IMAGE_CAPTION_EXTENSION: str = '.txt'

PATH_STATIC: Path = Path(__file__).parent / 'static'


def create_gallery(
    title: str,
    destination_path: Path,
    max_image_size: Dimension,
    max_thumbnail_size: Dimension,
    full_image_filenames: list[Path],
) -> Gallery:
    images = [create_image(image) for image in sorted(full_image_filenames)]
    images = list(link_images(images))

    return Gallery(
        title=title,
        destination_path=destination_path,
        max_image_size=max_image_size,
        max_thumbnail_size=max_thumbnail_size,
        images=images,
    )


def link_images(images: list[Image]) -> Iterator[Image]:
    """Assign the predecessor and successor for every image."""
    for previous_image, image, next_image in window(images):
        yield dataclasses.replace(
            image,
            previous_image=previous_image,
            next_image=next_image,
        )


def generate_gallery(gallery: Gallery, config: Config) -> None:
    # Create destination path if it doesn't exist.
    if not gallery.destination_path.exists():
        debug(
            'Destination path "{}" does not exist, creating it.',
            gallery.destination_path,
        )
        gallery.destination_path.mkdir()

    gallery = add_image_captions(gallery)

    if not config.html_only:
        generate_images(gallery, config.resize_images, config.optimize_images)

    render_html_pages(gallery)

    if not config.html_only:
        copy_additional_static_files(gallery.destination_path)

    debug('Done.')


def add_image_captions(gallery: Gallery) -> Gallery:
    captioned_images = [
        dataclasses.replace(image, caption=load_caption(image))
        for image in gallery.images
    ]

    return dataclasses.replace(gallery, images=captioned_images)


def copy_additional_static_files(destination_path: Path) -> None:
    if not PATH_STATIC.exists():
        debug(
            'Path "{}", does not exist; not copying any static files.',
            PATH_STATIC,
        )
        return

    filenames = list(sorted(PATH_STATIC.iterdir()))
    if not filenames:
        debug('No static files to copy.')
        return

    for filename in filenames:
        debug('Copying static file "{}" ...', filename)
        source = PATH_STATIC / filename
        destination = destination_path / filename
        shutil.copy(source, destination_path)


def window(
    images: Iterable[Image],
) -> Iterator[tuple[Optional[Image], Image, Optional[Image]]]:
    """Return a sliding window over the images.

    Returns a triple. Its first and third element can be `None`, but the
    second element never is.
    """
    n = 3
    padded_list: list[Optional[Image]] = list(images)
    padded_list.insert(0, None)
    padded_list.append(None)

    it = iter(padded_list)
    result = tuple(islice(it, n))

    if len(result) == n:
        yield result  # type: ignore

    for elem in it:
        result = result[1:] + (elem,)
        yield result  # type: ignore


def create_image(full_filename: Path) -> Image:
    basename = full_filename.stem
    extension = full_filename.suffix

    thumbnail_filename = f'{basename}_t{extension}'

    return Image(
        full_filename=full_filename,
        path=full_filename.parent,
        filename=full_filename.name,
        thumbnail_filename=thumbnail_filename,
        page_name=basename,
    )


def load_caption(image: Image) -> Optional[str]:
    """Load image caption from file."""
    filename = image.path / (image.filename + IMAGE_CAPTION_EXTENSION)
    return read_first_line(filename)


# -------------------------------------------------------------------- #


def handle_duplicate_filenames(paths: Iterable[Path]) -> None:
    duplicates = find_duplicate_filenames(paths)
    if not duplicates:
        return

    print('Found duplicate filenames:')
    for filename, paths in duplicates:
        print(f'  + "{filename}" appears in the following paths:')
        for path in paths:
            print(f'    - {path}')

    sys.exit('Clashing target filenames, aborting.')



"""
gallerize.models
~~~~~~~~~~~~~~~~

:Copyright: 2007-2021 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
from typing import Optional


@dataclass(frozen=True)
class Dimension:
    width: int
    height: int


@dataclass(frozen=True)
class Image:
    full_filename: Path
    path: Path
    filename: str
    thumbnail_filename: str
    page_name: str
    caption: Optional[str] = None
    previous_image: Optional[Image] = None
    next_image: Optional[Image] = None


@dataclass(frozen=True)
class Gallery:
    title: str
    destination_path: Path
    max_image_size: Dimension
    max_thumbnail_size: Dimension
    images: list[Image]


@dataclass(frozen=True)
class Config:
    html_only: bool
    resize_images: bool
    optimize_images: bool



"""
gallerize.files
~~~~~~~~~~~~~~~

Generic file utilities

:Copyright: 2007-2021 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from __future__ import annotations
from collections import defaultdict
from pathlib import Path
from typing import Iterable, Optional

from .logging import debug


def find_duplicate_filenames(
    paths: Iterable[Path],
) -> list[tuple[str, list[Path]]]:
    d = defaultdict(list)
    for path in paths:
        key = path.name.lower()
        d[key].append(path)

    return [item for item in d.items() if len(item[1]) > 1]


def read_first_line(filename: Path) -> Optional[str]:
    """Read the first line from the specified file."""
    try:
        text = filename.read_text(encoding='utf-8')
        first_line = text.splitlines()[0]
        return first_line.strip()
    except IOError:
        # File does not exist (OK) or cannot be read (not really OK).
        return None


def write_file(filename: Path, content: str) -> None:
    debug('Writing "{}" ...', filename)
    filename.write_text(content, encoding='utf-8')



"""
:Copyright: 2007-2021 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from pathlib import Path

import pytest

from gallerize.main import find_duplicate_filenames


@pytest.mark.parametrize(
    ('paths', 'expected'),
    [
        (
            [
                Path('dir1/foo.txt'),
                Path('dir1/bar.txt'),
                Path('foo.txt'),
                Path('dir2/Foo.tXt'),
            ],
            {
                'foo.txt': [
                    Path('dir1/foo.txt'),
                    Path('foo.txt'),
                    Path('dir2/Foo.tXt'),
                ],
            },
        ),
        (
            [
                Path('xy/one.jpg'),
                Path('xy/two.jpg'),
                Path('xy/three.jpg'),
                Path('one.jpg'),
                Path('three.jpg'),
            ],
            {
                'one.jpg': [Path('xy/one.jpg'), Path('one.jpg')],
                'three.jpg': [Path('xy/three.jpg'), Path('three.jpg')],
            },
        ),
    ],
)
def test_find_duplicate_filenames(paths, expected):
    assert dict(find_duplicate_filenames(paths)) == expected
